import json
import sqlite3
from pathlib import Path

from loguru import logger
from mircat_v2.configs import write_config, dbase_schema_file


def add_dbase_subparser(subparsers):
    # Add subcommands
    dbase_parser = subparsers.add_parser(
        "dbase",
        help="Database management commands",
        description="mircat-v2 database management commands.",
    )
    dbase_subparsers = dbase_parser.add_subparsers(
        dest="dbase_command", description="Database operations:"
    )

    # Add subcommand for creating a new database
    create_parser = dbase_subparsers.add_parser("create", help="Create a new database")
    create_parser.add_argument(
        "dbase_path",
        type=Path,
        help="Path to the database file",
    )
    create_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite the database if it already exists",
    )

    set_parser = dbase_subparsers.add_parser(
        "set", help="Set an existing sqlite database file to be used in mircat-v2"
    )
    set_parser.add_argument(
        "dbase_path", type=Path, help="Path to the existing database file."
    )
    dbase_subparsers.add_parser(
        "update", help="Update the database schema to the newest version."
    )

    # add subcommand for querying the database
    query_parser = dbase_subparsers.add_parser("query", help="Query the database")
    query_parser.add_argument(
        "query",
        type=str,
        help="SQL query to execute",
    )


def run_dbase_command(args) -> None:
    """Run the appropriate function depending on the given dbase command
    Parameters:
        args: The arguments passed to the CLI
    """
    match args.dbase_command:
        case "create":
            create_database(args.dbase_path, args.overwrite)
        case "set":
            set_database(args.dbase_path)
        case "update":
            logger.info("update command not functional yet")
        case "query":
            logger.info("query command not functional yet.")


def create_database(dbase_path: Path, overwrite: bool = False) -> None:
    """
    Create a new SQLite database at the specified path.

    Parameters:
        dbase_path (Path): Path to the database file.
        overwrite (bool): If True, overwrite the database if it already exists.
    """
    # Raise an error if the database exists and we don't want to overwrite
    if dbase_path.exists() and not overwrite:
        raise FileExistsError(
            f"Database file {dbase_path} already exists. Use `mircat-v2 dbase set` to use it or `--overwrite` to completely overwrite."
        )
    # Confirm that you want to delete the previous database
    elif dbase_path.exists() and overwrite:
        response = input(
            "Are you sure you want to overwrite the existing database? This will not copy data to the new one. (y/n): "
        )
        if response.lower() != "y":
            logger.info(f"Left existing database at {dbase_path} intact.")
            return
        logger.info(f"Overwriting existing database at {dbase_path}")
        dbase_path.unlink()
    # Create the database
    with sqlite3.connect(dbase_path) as conn:
        create_tables_from_schema(conn)
    logger.success(f"Database created at {dbase_path}")
    conn.close()
    save_dbase_path(dbase_path)


def set_database(dbase_path: Path) -> None:
    """Set the database at the given path to be used for database operations"""
    if not dbase_path.exists():
        raise FileNotFoundError(
            f"The sqlite database at {dbase_path} was not found. If you want to create one here, please use `mircat-v2 dbase create {dbase_path}`"
        )
    # The sql create script only makes tables if they do not exist - so it's safe to run this
    with sqlite3.connect(dbase_path) as conn:
        create_tables_from_schema(conn)
    conn.close()
    save_dbase_path(dbase_path)
    logger.success("Set {} as the database for mircat-v2.", dbase_path)


# TODO - write the update option using the default dbase schema that we store to the config.
def save_dbase_path(dbase_path: Path) -> None:
    dbase_config = str(dbase_path.resolve())
    write_config(dbase_config, "dbase", "dbase_path")


def create_tables_from_schema(conn: sqlite3.Connection) -> None:
    cursor = conn.cursor()
    # Run the creation script
    with dbase_schema_file.open("r") as f:
        schema_script = f.read()
    cursor.executescript(schema_script)
    conn.commit()


def insert_data_batch(dbase_config: dict, table: str, data_records: list[dict]):
    """Insert a batch of data into the database for a specific table."""
    if not data_records:
        return
    dbase_path = Path(dbase_config["dbase_path"])
    conn = sqlite3.connect(dbase_path)
    cursor = conn.cursor()
    try:
        # Get the columns that exist in the table
        table_schema = dbase_config["tables"].get(table)
        if not table_schema:
            raise ValueError(
                f"Table {table} does not exist in the standard MirCAT database schema."
            )
        # Check if the table exists
        cursor.execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table}'"
        )
        if not cursor.fetchone():
            logger.info(f"Table {table} does not exist. Creating it.")
            create_tables_from_schema(conn)
            logger.info(f"Table {table} created successfully.")
        table_columns = list(table_schema.keys())

        # Process each record in the batch
        insert_data = []
        for metadata in data_records:
            # Filter metadata to only include columns that exist in the table and
            # put in correct order. We use get to insert null values for missing columns
            insert_data.append(
                {
                    col: metadata.get(col)
                    for col in table_columns
                    if col != "PRIMARY KEY"
                }
            )
        logger.debug("Insert data: {}", insert_data)

        if insert_data:
            # Use the first record to determine column order
            columns = list(insert_data[0].keys())
            placeholders = ", ".join(["?" for _ in columns])
            insert_sql = (
                f"REPLACE INTO {table} ({', '.join(columns)}) VALUES ({placeholders})"
            )

            # Prepare values for batch insert
            values_list = []
            for metadata in insert_data:
                values = [metadata.get(col) for col in columns]
                values_list.append(values)

            # Execute batch insert
            cursor.executemany(insert_sql, values_list)
            conn.commit()
            logger.success(f"Inserted {len(values_list)} records into {table} table")
    except Exception as e:
        conn.rollback()
        logger.error(f"Error inserting batch: {e}")
        raise
    finally:
        conn.close()



import json
import pprint
from argparse import ArgumentParser
from pathlib import Path
from shutil import copytree

from loguru import logger
from mircat_v2.configs import write_config, read_models_config

models_path = Path(__file__).parent.resolve()


def add_models_subparser(subparsers: ArgumentParser) -> None:
    """Add the model subcommand to mircat-v2 cli interface
    Parameters:
        subparsers: ArgumentParser - the main parser from mircat_v2.main
    """
    # Model specific operations
    models_parser = subparsers.add_parser(
        "models",
        description="""List models available, copy new models to the correct mircat-v2 location, or update the config file for segmentation""",
        help="List, add or update models available to mircat-v2.",
    )
    models_subparser = models_parser.add_subparsers(dest="models_command")
    list_subparser = models_subparser.add_parser(
        "list",
        help="List tasks and their descriptions currently available to mircat-v2.",
    )
    list_subparser.add_argument(
        "-t",
        "--task",
        type=str,
        default="all",
        help="Specific task number to list configuration for. Default is to show all tasks.",
    )
    add_command = models_subparser.add_parser(
        "add",
        description="Add nnUNet model(s) to mircat-v2 options.",
        help="Add nnUNet model or folder of nnUNet models to mircat library",
    )
    add_command.add_argument(
        "folder",
        type=Path,
        help="Path to nnUNet model or folder containing multiple nnUNet models. \
            Should either be in the format DatasetXXX_* or be a folder containing folders in that format.",
    )
    add_command.add_argument(
        "--overwrite",
        action="store_true",
        help="If a model task already exists in the mircat models file, overwrite it.",
    )
    models_subparser.add_parser("update", help="Update the model configurations file.")


def _get_available_models():
    task_folders = sorted(models_path.glob("Dataset*/"))
    logger.debug(
        """Found {} models in {}. They are:\n{}""",
        len(task_folders),
        str(models_path),
        pprint.pformat([str(x.name) for x in task_folders]),
    )
    if len(task_folders) == 0:
        raise FileNotFoundError(
            f"No folders found in {models_path}. Use mircat-v2 models copy to copy in some nnUNet models!"
        )
    return task_folders


def update_models_config():
    # We call them tasks as this is the nnUNet style - Dataset###_descriptor
    task_folders = _get_available_models()
    # The keys here are what UNet calls each model type
    # The values are what we will use as flags to select a model
    unet_configs = {
        "2d": "2d",
        "3d_fullres": "3d",
        "3d_lowres": "3d_lowres",
        "3d_cascade_fullres": "3d_cascade",
    }
    model_configs = {}
    for folder in task_folders:
        splits = folder.name.split("_")
        task = splits[0].replace("Dataset", "")
        description = "_".join(splits[1:])
        task_models = sorted(folder.glob("*/"))
        logger.debug(f"Task: {task}, Models: {task_models}")
        task_config = {"description": description, "models": {}}
        if len(task_models) == 0:
            logger.warning(f"No model options found in {folder}.")
            continue
        # Pull a reference model
        ref_model = task_models[0]
        if not (ref_model / "dataset.json").exists():
            logger.error(f"No dataset info found for {folder}. Skipping.")
            continue
        if not (ref_model / "plans.json").exists():
            logger.error(f"No model plans found for {folder}. Skipping.")
        with (ref_model / "dataset.json").open() as f:
            dataset = json.load(f)
        with (ref_model / "plans.json").open() as f:
            plans = json.load(f)
            configs = plans["configurations"]
        labels = dataset.get("labels")
        task_config["labels"] = labels
        # Get all of the nnUNet model options for the task that are available.
        for unet_config, option in unet_configs.items():
            for model in task_models:
                if unet_config in model.name:
                    model_specific_config = configs.get(unet_config)
                    task_config["models"][option] = {
                        "path": str(model),
                        "patch_size": model_specific_config.get("patch_size"),
                        "spacing": model_specific_config.get("spacing"),
                    }
        model_configs[task] = task_config
    write_config(model_configs, "models")


def add_models_to_mircat(args) -> None:
    folder: Path = args.folder
    overwrite: bool = args.overwrite
    if overwrite:
        confirmed = input(
            "Overwrite option was given to copy command. Are you sure you want to overwrite existing models? (y/n) "
        )
        confirmed = confirmed == "y"
        if not confirmed:
            logger.info("Aborting copy operation.")
            return

    folder = folder.resolve()
    if not folder.is_dir():
        logger.error(
            f"Given path {folder} is not a folder. Please make sure you are passing an nnUNet folder."
        )
        return
    if not any(folder.iterdir()):
        logger.error(
            f"Folder {folder} is empty. Please check that your path is correct."
        )
        return
    logger.info(f"Copying models from {folder} to {models_path}.")
    if folder.name.startswith("Dataset"):
        logger.info(f"Found 1 model to copy: {folder}")
        logger.info(f"Copying {folder} to {models_path}")
        destination = models_path / folder.name
        copytree(folder, destination, dirs_exist_ok=overwrite)
        logger.success(
            f"{folder} succesfully copied to mircat-v2 models. Updating config."
        )
    else:
        models = list(folder.glob("*/Dataset*"))
        n_models = len(models)
        if n_models == 0:
            logger.error(
                f"No models found in {folder}. Please make sure the input is either of format 'DatasetXXX_*' or folder/DatasetXXX_*"
            )
            return
        model_tasks = {
            model.name.split("_")[0].replace("Dataset", ""): model for model in models
        }
        logger.info(
            f"Found {n_models} models to copy in {folder} - Task numbers are {', '.join(model_tasks.keys())}"
        )
        for i, (task, model_path) in enumerate(model_tasks.items(), start=1):
            destination = models_path / model_path.name
            copytree(model_path, destination, dirs_exist_ok=overwrite)
            logger.success(
                f"{i}/{n_models} - Succesfully copied task {task} from {model_path} to mircat library."
            )
        logger.info("All models copied. Updating config.")
    update_models_config()


def list_mircat_models(args):
    configs = read_models_config()
    if args.task == "all":
        pprint.pprint(configs, indent=2, sort_dicts=False)
        return
    else:
        subdict = configs.get(args.task)
        if subdict is None:
            logger.error(
                f"Task number {args.task} not found in model configs. Available tasks are\n\t{list(configs.keys())}",
            )
            return
        pprint.pprint(subdict, indent=2, sort_dicts=False)



import sys
import json
import pprint

from math import floor
from multiprocessing import cpu_count
from pathlib import Path

from loguru import logger

from mircat_v2.configs import (
    read_models_config,
    read_dbase_config,
    set_threads_per_process,
)
from mircat_v2.nifti import load_nifti_paths
from .models import models_path

lib_path = Path(__file__).parent.resolve()


def add_segmentation_subparser(subparsers):
    """Add segmentation specific subcommands to mircat-v2 cli interface
    :param subparsers: the subparser for CLI
    """
    # Add subcommands for segmentation itself
    seg_parser = subparsers.add_parser(
        "segment",
        description="""Segment nifti files using nnUNet - accelerated by GPU if available. \
            Input files will be resampled and stored in temporary files to accelerate nnUNet. \
            Make sure you have enough disk space for resampled images!""",
        help="Segment nifti files using nnUNet models.",
    )
    seg_parser.add_argument(
        "niftis",
        type=Path,
        help="Path to a nifti file or a text file containing a list of nifti files",
    )
    seg_parser.add_argument(
        "--data-dir",
        "-d",
        type=Path,
        default=Path(""),
        help="Data directory for the input niftis (i.e. {data_dir}/nifti_path) - useful for remapping in docker. Default is no data directory.",
    )
    seg_parser.add_argument(
        "-tl",
        "--task-list",
        type=str,
        nargs="+",
        required=True,
        help="Space separated list of segmentation model tasks to perform. Identified by nnUNet dataset number. Use mircat-v2 models list to see all available tasks.",
    )
    seg_parser.add_argument(
        "-m",
        "--model-types",
        type=str,
        nargs="+",
        default=["3d"],
        choices=["2d", "3d", "3d_lowres", "3d_cascade"],
        help="Default = 3d for all tasks. The nnUNet model subtype to run for each task, space separated.",
    )
    seg_parser.add_argument(
        "--device",
        "-dv",
        type=str,
        default="cuda:0",
        help="Device for pytorch/nnunet to use for segmentation. Default is cuda:0",
    )
    seg_parser.add_argument(
        "-n",
        "--n-processes",
        type=int,
        default=1,
        help="Default = 1. Number of processes to use for pre/post processing the images.",
    )
    seg_parser.add_argument(
        "-t",
        "--threads-per-process",
        type=int,
        default=4,
        help="Default = 4. Maximum number of threads each process should use.",
    )
    seg_parser.add_argument(
        "-c",
        "--cache-size",
        type=int,
        default=10,
        help="The number of nifti files to work on at one time. Default = 10. This includes saving preprocessed files to disk, so be mindful of storage.",
    )
    seg_parser.add_argument(
        "-db",
        "--dbase-insert",
        action="store_true",
        help="Store model results (completed/failed for each nifti) in the mircat-v2 database. Must be setup!",
    )
    seg_parser.add_argument(
        "-ir",
        "--image-resampler",
        type=str,
        choices=["lanczos", "bspline", "gaussian", "linear"],
        default="bspline",
        help="Interpolator for resampling original images. Default = bspline (speed and quality balance)",
    )
    seg_parser.add_argument(
        "-lr",
        "--label-resampler",
        type=str,
        choices=["gaussian", "linear", "nearest"],
        default="gaussian",
        help="Interpolator for resampling segmentation images. Default = gaussian (slowest but best)",
    )


def segment_nifti_files(args):
    try:
        from mircat_v2.segmentation.segmentor import MircatSegmentor
    except ModuleNotFoundError as e:
        logger.error(
            "Could not import the segmentation module. Please make sure you have installed mircat-v2 with `pip install mircat-v2[seg]`."
        )
        raise e
    logger.debug(
        "Starting segmentation process with the following args:\n{}",
        pprint.pformat(args),
    )
    task_configs = read_models_config()
    _validate_segmentation_args(task_configs, args)
    niftis = load_nifti_paths(args.niftis, args.data_dir)
    segmentor = MircatSegmentor(
        niftis=niftis,
        task_list=args.task_list,
        model_types=args.model_types,
        device=args.device,
        task_configs=task_configs,
        n_processes=args.n_processes,
        threads_per_process=args.threads_per_process,
        cache_size=args.cache_size,
        dbase_config=args.dbase_config,
        img_resampler=args.image_resampler,
        lbl_resampler=args.label_resampler,
    )
    segmentor.run()


def _validate_segmentation_args(task_configs: dict, args):
    """Internal function to validate the given arguments for nnUNet segmentation
    :param model_configs: the internal configuration file for mircat-v2 loaded as a dictionary
    :param args: the passed input parameters
    """
    # Make sure the input argument for the nifti file(s) exists
    args.niftis = args.niftis.resolve()
    if not args.niftis.exists():
        logger.error(
            f"The input nifti file/list of files {args.niftis} does not exist. Please double check your paths."
        )
        sys.exit(1)
    available_models = list(task_configs.keys())
    # Make sure all tasks are available
    if not all([task in available_models for task in args.task_list]):
        missing_tasks = [
            task for task in args.task_list if task not in available_models
        ]
        logger.error(
            "The following tasks are missing from mircat-v2 config file {}. Please use `mircat-v2 models add` to place them in the correct location.",
            missing_tasks,
        )
        sys.exit(1)
    # If only one model type was given, apply it for all tasks
    if len(args.model_types) == 1:
        model_type = args.model_types[0]
        logger.debug(
            "One model type passed to --model-types. Will apply for all given tasks."
        )
        args.model_types = [model_type for task in args.task_list]
    # Ensure that each task has a specified model type
    if len(args.model_types) != len(args.task_list):
        logger.error(
            f"Number of model type parameters must match the number of tasks. Currently have {len(args.model_types)} model types and {len(args.task_list)} tasks.\n\t\
            Please either specify one model type [3d, 2d, 3d_lowres, 3d_casade] for all tasks or have a specific model type for each task."
        )
        sys.exit(1)
    # Ensure that each task has weights for the specific model type
    for task, model in zip(args.task_list, args.model_types):
        model_weights = task_configs[task]["models"].get(model, dict())
        logger.debug(f"{task} configurations: {model_weights}")
        if not model_weights:
            logger.error(f"No {model} configuration found for task {task}.")
            sys.exit(1)
    # Check if the mircat-v2 dbase exists.
    if args.dbase_insert:
        args.dbase_config = read_dbase_config()
    else:
        args.dbase_config = {}

    set_threads_per_process(args)



import os
import sys

os.environ["nnUNet_raw"] = ""
os.environ["nnUNet_preprocessed"] = ""
os.environ["nnUNet_results"] = ""

import argparse

from loguru import logger

from mircat_v2.configs import logger_setup
from mircat_v2.dicom_conversion import (
    convert_dicoms_to_nifti,
    add_dicom_conversion_subparser,
)
from mircat_v2.dbase import add_dbase_subparser
from mircat_v2.segmentation import add_segmentation_subparser
from mircat_v2.segmentation.models import (
    add_models_subparser,
    update_models_config,
    add_models_to_mircat,
    list_mircat_models,
)
from mircat_v2.stats import add_stats_subparser

__version__ = "0.1.1"


def main() -> None:
    ## We have to do local imports so that optional dependencies are not loaded
    """Mirshahi Lab CT Analysis Toolkit (v2) CLI entry point."""
    parser = argparse.ArgumentParser(
        description="Mirshahi Lab CT Analysis Toolkit (v2)"
    )
    verbose_parser = parser.add_mutually_exclusive_group()
    verbose_parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Write debugging logs in the terminal.",
    )
    verbose_parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Reduce output logs in the terminal to only success and error messages.",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)
    # Add all the subparsers
    add_dicom_conversion_subparser(subparsers)
    add_dbase_subparser(subparsers)
    add_models_subparser(subparsers)
    add_segmentation_subparser(subparsers)
    add_stats_subparser(subparsers)
    # Parse the command line arguments
    args = parser.parse_args()
    logger_setup(args.verbose, args.quiet)
    # Handle the commands
    # Convert dicoms to nifti
    if args.command == "convert":
        convert_dicoms_to_nifti(args)

    # Perform a database operation
    elif args.command == "dbase":
        try:
            from mircat_v2.dbase import run_dbase_command
        except ImportError as e:
            logger.error(
                "Database functionality requires the 'dbase' extra. "
                "Please install with: pip install mircat-v2[dbase]"
            )
            exit(1)
        run_dbase_command(args)

    # Segment nifti files using a model
    elif args.command == "segment":
        try:
            from mircat_v2.segmentation import segment_nifti_files
        except ImportError as e:
            logger.error(
                "Segmentation functionality requires the 'seg' extra. "
                "Please install with: pip install mircat-v2[seg]"
            )
            exit(1)
        segment_nifti_files(args)

    # Look into segmentation models
    elif args.command == "models":
        if args.models_command == "list":
            list_mircat_models(args)
        elif args.models_command == "update":
            update_models_config()
        elif args.models_command == "add":
            add_models_to_mircat(args)

    # Run stats on niftis with segmentations
    elif args.command == "stats":
        try:
            from mircat_v2.stats import run_stats
        except ImportError as e:
            logger.error(
                "Stats functionality requires the 'stats' extra. "
                "Please install with: pip install mircat-v2[stats]"
            )
            exit(1)
        run_stats(args)
    else:
        logger.error(f"Unknown command: {args.command}")
        parser.print_help()
        exit(1)



import pickle
import numpy as np
from pathlib import Path


def predict_contrast(intensities: dict):
    """Predict CT contrast phase based on intensity statistics.
    :param intensities: Dictionary containing volume and intensity statistics.
    """
    # These are the structures we use for contrast prediction
    structures = [
        "liver",
        "pancreas",
        "bladder",
        "gallbladder",
        "heart",
        "aorta",
        "inferior_vena_cava",
        "portal_vein_and_splenic_vein",
        "iliac_vena_left",
        "iliac_vena_right",
        "iliac_artery_left",
        "iliac_artery_right",
        "pulmonary_vein",
        "brain",
        "colon",
        "small_bowel",
        "internal_carotid_artery_right",
        "internal_carotid_artery_left",
        "internal_jugular_vein_right",
        "internal_jugular_vein_left",
    ]
    features = [intensities.get(f"{structure}_avg_hu", 0.0) for structure in structures]
    models_file = Path(__file__).parent / "contrast_phase_classifiers_2024_07_19.pkl"
    with models_file.open("rb") as f:
        models = pickle.load(f)
    preds = []
    for clf in models.values():
        pred = clf.predict([features])[0]
        preds.append(pred)
    preds = np.array(preds)
    pi_time = round(float(np.mean(preds)), 2)
    pi_time_std = round(float(np.std(preds)), 4)
    phase, probability = pi_time_to_phase(pi_time)
    return {
        "pi_time": pi_time,
        "pi_time_std": pi_time_std,
        "phase": phase,
        "probability": probability,
    }


def pi_time_to_phase(pi_time: float) -> str:
    """
    Convert the pi time to a phase and get a probability for the value.

    native: 0-10
    arterial_early: 10-30
    arterial_late:  30-60
    portal_venous:  60-100
    delayed: 100+

    returns: phase, probability
    """
    if pi_time < 5:
        return "native", 1.0
    elif pi_time < 10:
        return "native", 0.7
    elif pi_time < 20:
        return "arterial_early", 0.7
    elif pi_time < 30:
        return "arterial_early", 1.0
    elif pi_time < 50:
        return "arterial_late", 1.0
    elif pi_time < 60:
        return "arterial_late", 0.7  # in previous version: "portal_venous"
    elif pi_time < 70:
        return "portal_venous", 1.0
    elif pi_time < 90:
        return "portal_venous", 1.0
    elif pi_time < 100:
        return "portal_venous", 0.7
    else:
        return "portal_venous", 0.3



import SimpleITK as sitk
import numpy as np

from loguru import logger
from mircat_v2.stats.utilities import calculate_shape_stats


def calculate_vertebrae_stats(nifti, seg_id: str, label_map: dict) -> dict:
    """Calculate vertebrae midlines and height statistics for a given nifti file
    :param nifti: StatsNifti object containing the nifti file and its metadata.
    :param seg_id: The segmentation ID to calculate statistics for.
    :param label_map: Dictionary mapping structure names to label IDs.
    :return: Dictionary containing vertebrae midlines and height statistics.
    """
    vertebrae_labels = {
        k.replace("vertebrae_", ""): v
        for k, v in label_map.items()
        if "vertebrae" in k.lower()
    }
    segmentation = nifti.id_to_seg_map[seg_id]
    shape_stats = calculate_shape_stats(segmentation)
    vertebrae_stats = {}
    previous_midline = 1e6
    correct_vertebrae_order = 1
    for vertebra, label in vertebrae_labels.items():
        if label in shape_stats.GetLabels():
            vertebra_indicies = shape_stats.GetIndexes(label)
            z_indices = vertebra_indicies[2::3]
            midline = int(np.median(z_indices))
            vertebrae_stats[vertebra] = {"midline": midline}
            if midline > previous_midline:
                logger.warning(
                    f"Vertebrae {vertebra} midline {midline} is greater than previous midline {previous_midline}. "
                    "This may indicate an incorrect order of vertebrae."
                )
                correct_vertebrae_order = 0
            previous_midline = midline

    lowest_vertebra = max(vertebrae_stats, key=lambda k: vertebrae_stats[k]["midline"])
    highest_vertebra = min(vertebrae_stats, key=lambda k: vertebrae_stats[k]["midline"])
    vertebrae_stats["lowest_vertebra"] = lowest_vertebra
    vertebrae_stats["highest_vertebra"] = highest_vertebra
    vertebrae_stats["correct_vertebrae_order"] = correct_vertebrae_order
    vertebrae_stats["abdominal_scan"] = int(
        "L1" in vertebrae_stats and "L5" in vertebrae_stats
    )
    vertebrae_stats["chest_scan"] = int(
        "T3" in vertebrae_stats and "T12" in vertebrae_stats
    )
    return vertebrae_stats


# TODO - Add vertebrae height statistics



import SimpleITK as sitk
from pathlib import Path
from loguru import logger
from multiprocessing import Pool
from threadpoolctl import threadpool_limits

from mircat_v2.configs import read_dbase_config, read_models_config
from mircat_v2.dbase import insert_data_batch
from mircat_v2.nifti import StatsNifti
from mircat_v2.stats.aorta import calculate_aorta_stats
from mircat_v2.stats.contrast import predict_contrast
from mircat_v2.stats.tissues import calculate_tissue_stats
from mircat_v2.stats.vertebrae import calculate_vertebrae_stats
from mircat_v2.stats.vol_int import calculate_volume_and_intensity_stats


class Analyzer:
    """Analyzer class for processing and analyzing tasks with specified parameters."""

    # Map for resolution settings
    resolution_map = {
        "normal": [1.0, 1.0, 1.0],
        "high": [0.75, 0.75, 0.75],
        "highest": [0.5, 0.5, 0.5],
    }
    # Segmentations needed for specific tasks in order of preference
    task_map = {
        "vol_int": ["999", "total"],
        "contrast": ["999", "total"],
        "vertebrae": ["999", "total"],
        "aorta": ["999", "total"],
        "tissues": ["485", "481", "tissues"],
        "body": ["299", "300", "body"],
    }

    empty_stats_insert = {
        "metadata": [],
        "vol_int": [],
        "contrast": [],
        "vertebrae": [],
        "aorta_metrics": [],
        "aorta_diameters": [],
        "tissues_volumetric": [],
        "tissues_vertebral": [],
    }

    def __init__(
        self,
        niftis: list[Path],
        task_list: list[str],
        resolution: str,
        image_resampler: str,
        label_resampler: str,
        n_processes: int,
        threads_per_process: int,
        dbase_insert: bool = False,
        overwrite: bool = False,
        gzip: bool = False,
    ):
        self.niftis = niftis
        self.total_niftis = len(niftis)
        if task_list == ["all"]:
            task_list = list(self.task_map.keys())
        # This orders the tasks by the order they are defined in task_map
        self.task_list = [task for task in self.task_map if task in task_list]
        if "tissues" in self.task_list and "body" not in task_list:
            self.task_list.append("body")
        self.resolution = self.resolution_map.get(resolution)
        self.image_resampler = image_resampler
        self.label_resampler = label_resampler
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        # Read database configuration if dbase_insert is True
        self.dbase_insert = dbase_insert
        if dbase_insert:
            self.dbase_config = read_dbase_config()
        else:
            self.dbase_config = {}
        # Read segmentation configuration
        seg_config = read_models_config()
        # We only need the labels for each task
        self.segmentation_labels = {
            task: seg_config[task]["labels"] for task in seg_config
        }
        self.overwrite = overwrite
        self.gzip = gzip

    def analyze(self, nifti_path: Path | str) -> StatsNifti:
        """Analyze a single nifti file.
        :param nifti_path: Path to the nifti file to analyze.
        """
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        with threadpool_limits(self.threads_per_process):
            nifti = StatsNifti(nifti_path, self.overwrite)
            # Preprocess the nifti file for statistics
            (
                nifti.check_for_segmentations(
                    self.task_list, self.task_map
                ).preprocess_for_stats(
                    self.resolution, self.image_resampler, self.label_resampler
                )
            )
            self.analyze_tasks(nifti)
        nifti.save_json_stats(self.gzip)
        if self.dbase_insert:
            nifti.format_stats_for_db()
        return nifti

    def analyze_tasks(self, nifti: StatsNifti) -> None:
        for task in self.task_list:
            seg_id = nifti.task_to_id_map[task]
            seg_labels = self.segmentation_labels[seg_id]
            try:
                match task:
                    case "vol_int":
                        vol_int_stats = calculate_volume_and_intensity_stats(
                            nifti, seg_id, seg_labels
                        )
                        nifti.add_stats("vol_int", vol_int_stats)
                    case "contrast":
                        if not nifti.stats.get("vol_int", {}):
                            logger.warning(
                                "Organ intensities are needed for contrast prediction. Skipping contrast task."
                            )
                            continue
                        contrast_stats = predict_contrast(nifti.stats["vol_int"])
                        nifti.add_stats("contrast", contrast_stats)
                    case "vertebrae":
                        vertebrae_stats = calculate_vertebrae_stats(
                            nifti, seg_id, seg_labels
                        )
                        # Add the structural vert stats to the metadata for easy searching
                        for key in [
                            "lowest_vertebra",
                            "highest_vertebra",
                            "correct_vertebrae_order",
                            "abdominal_scan",
                            "chest_scan",
                        ]:
                            nifti.add_stats(
                                key, vertebrae_stats.pop(key, None), "metadata"
                            )
                        nifti.add_stats("vertebrae", vertebrae_stats)
                        nifti.set_vertebrae_midlines(vertebrae_stats)
                    case "aorta":
                        aorta_stats = calculate_aorta_stats(nifti, seg_id, seg_labels)
                        nifti.add_stats("aorta", aorta_stats)
                    case "tissues":
                        # Need body segmentation for tissue stats
                        body_seg_id = nifti.task_to_id_map.get("body", None)
                        if not body_seg_id:
                            logger.warning(
                                "Body segmentation is required for tissue statistics. Skipping tissue task."
                            )
                            continue
                        body_labels = self.segmentation_labels[body_seg_id]
                        tissue_stats = calculate_tissue_stats(
                            nifti, seg_id, seg_labels, body_seg_id, body_labels
                        )
                        nifti.add_stats("tissues", tissue_stats)
            except Exception as e:
                logger.error(f"Error calculating {task} statistics: {e}")

    def run(self, batch_insert_size: int = 100) -> None:
        """Run the analysis on the provided NiFTi files.
        :param niftis: Path to a nifti file with mircat-v2 segmentations or a text file with a list of multiple mircat-v2 nifti files.
        """
        # self._get_nifti_list(niftis)
        if self.n_processes > 1:
            logger.info(f"Running analysis with {self.n_processes} processes.")
        else:
            logger.info("Running analysis with single process.")
        logger.info(
            f"Using resolution: {self.resolution} mm, image resampler: {self.image_resampler}, label resampler: {self.label_resampler}."
        )
        logger.info(f"Tasks to run: {self.task_list}.")
        logger.info(f"Starting analysis on {self.total_niftis} nifti files.")
        batch_data = self.empty_stats_insert.copy()
        with Pool(processes=self.n_processes) as pool:
            for i, nifti in enumerate(
                pool.imap_unordered(self.analyze, self.niftis), start=1
            ):
                logger.success(
                    f"Stats: [{i}/{self.total_niftis}] - Successfully calculated statistics for {nifti.name}."
                )
                if self.dbase_insert:
                    for key, list_data in nifti.db_stats.items():
                        batch_data[key].extend(list_data)
                    if i % batch_insert_size == 0:
                        logger.info(
                            "inserting data from {} nifti files into the database.",
                            batch_insert_size,
                        )
                        for key, list_data in batch_data.items():
                            if list_data:
                                insert_data_batch(self.dbase_config, key, list_data)
                        batch_data = self.empty_stats_insert.copy()

        if self.dbase_insert and batch_data:
            logger.info("inserting remaining nifti files into the database.")
            for key, list_data in batch_data.items():
                if list_data:
                    insert_data_batch(self.dbase_config, key, list_data)



from loguru import logger

from mircat_v2.stats.utilities import (
    calculate_2d_areas,
    calculate_3d_volumes,
    calculate_3d_intensities,
)


def calculate_tissue_stats(
    nifti,
    tissue_seg_id: str,
    tissue_label_map: dict,
    body_seg_id: str,
    body_label_map: dict,
) -> dict:
    """Calculate tissue statistics for a given Nifti object.

    Parameters:
        nifti (StatsNifti): The Nifti object containing the image data.
        tissue_seg_id (str): The segmentation ID for the tissue.
        tissue_label_map (dict): A mapping of tissue labels to their IDs.
        body_seg_id (str): The segmentation ID for the body.
        body_label_map (dict): A mapping of body labels to their IDs.
    Returns:
        dict: A dictionary containing the calculated tissue statistics.
    """
    if not nifti.vertebrae_midlines:
        logger.error(
            "Vertebrae midlines are not set. Cannot calculate tissue statistics."
        )
        return {}
    logger.debug(f"Calculating tissue statistics for {nifti.name}")
    # Get the data points we need
    stats = {"volumetric": {}, "vertebral": {}}
    tissue_seg = nifti.id_to_seg_map.get(tissue_seg_id)
    body_seg = nifti.id_to_seg_map.get(body_seg_id)
    vertebrae_midlines = nifti.vertebrae_midlines
    # We include the thoracic to measure around the heart
    verts_to_measure = {
        v: vertebrae_midlines[v]
        for v in ["T5", "T6", "T7", "T8", "L1", "L3", "L5"]
        if v in vertebrae_midlines
    }
    is_abdominal = "L1" in vertebrae_midlines and "L5" in vertebrae_midlines
    is_chest = "T3" in vertebrae_midlines and "T12" in vertebrae_midlines
    logger.debug(
        f"is_abdominal: {is_abdominal}, is_chest: {is_chest}, verts_to_measure: {verts_to_measure}"
    )

    def measure_tissues(endpoints=(0, None)):
        """Measure the tissue segmentation in a volumetric region."""
        tissue_volumes = calculate_3d_volumes(
            tissue_seg, tissue_label_map, endpoints=endpoints, unit="cm3"
        )
        tissue_intensities = calculate_3d_intensities(
            nifti.img, tissue_seg, tissue_label_map, endpoints=endpoints
        )
        body_volumes = calculate_3d_volumes(
            body_seg, body_label_map, endpoints=endpoints, unit="cm3"
        )
        body_intensities = calculate_3d_intensities(
            nifti.img, body_seg, body_label_map, endpoints=endpoints
        )
        merged_tissues = tissue_volumes.copy()
        for key, value in tissue_intensities.items():
            merged_tissues[key].update(value)
        merged_body = body_volumes.copy()
        for key, value in body_intensities.items():
            merged_body[key].update(value)
        # This final dictionary will be in the format: {structure: {stat: value}}
        return {**merged_tissues, **merged_body}

    # Measure the volumes and intensities of the body and tissue segmentation
    total_stats = measure_tissues()
    stats["volumetric"]["total"] = total_stats
    # Measure region specific volumes
    if is_abdominal:
        endpoints = (vertebrae_midlines["L1"], vertebrae_midlines["L5"])
        abd_stats = measure_tissues(endpoints=endpoints)
        stats["volumetric"]["abdominal"] = abd_stats
    if is_chest:
        endpoints = (vertebrae_midlines["T3"], vertebrae_midlines["T12"])
        chest_stats = measure_tissues(endpoints=endpoints)
        stats["volumetric"]["chest"] = chest_stats
    # Measure the stats for each vertebra
    for vertebra, midline in verts_to_measure.items():
        body_vert_stats = calculate_2d_areas(
            body_seg, body_label_map, midline, get_perimeter=True, units="cm"
        )
        tissue_vert_stats = calculate_2d_areas(
            tissue_seg, tissue_label_map, midline, get_perimeter=False, units="cm"
        )
        stats["vertebral"][vertebra] = {
            **body_vert_stats,
            **tissue_vert_stats,
        }
    return stats



import json
import traceback
from datetime import date
from pathlib import Path
from multiprocessing import Pool
from shutil import copyfile

import dicom2nifti
import dicom2nifti.settings as settings
from dicom2nifti.exceptions import ConversionError
from dicom2nifti.convert_dir import _remove_accents
import numpy as np
from pydicom import dcmread
from loguru import logger
from threadpoolctl import threadpool_limits
from mircat_v2.configs import read_dbase_config
from mircat_v2.dbase import insert_data_batch


class DicomConverter:
    """Class to convert DICOM files to NIfTI format using the dicom2nifti library.
    This class provides options for resampling, validation, and parallel processing.
    The output NIfTI files are saved in a specified directory with subdirectories based on the input DICOM file's metadata.
    """

    def __init__(
        self,
        axial_only: bool = False,
        no_mip: bool = False,
        resample: bool = True,
        resample_spline_interpolation_order: int = 3,
        resample_padding: int = -1024,
        validate_orthogonal: bool = False,
        validate_slice_increment: bool = True,
        validate_instance_number: bool = True,
        validate_slice_count: bool = True,
        n_processes: int = 1,
        threads_per_process: int = 4,
        min_slice_count: int = 30,
        db_batch_size: int = 100,
    ):
        """
        Initialize the DicomConverter with optional resampling settings.

        Args:
            axial_only (bool): Whether to only convert axial series. Default is False.
            no_mip (bool): Whether to skip converting MIP series. Default is False.
            resample (bool): Whether to resample the DICOM images to standard LAS orientation. Default is True.
            resample_spline_interpolation_order (int): The order of the spline interpolation for resampling. Default is 3.
            resample_padding (int): The padding value for resampling. Default is -1024 (Hounsfield units).
            validate_orthogonal (bool): Whether to validate the orthogonality of the DICOM images. Default is False.
                If True, gantry tilted images will not be converted.
            n_processes (int): The number of processes to use for conversion. Default is 1.
            threads_per_process (int): The maximum number of threads to use for conversion in each process. Default is 4.
            min_file_count (int): The minimum number of DICOM files in a folder required to perform conversion. Default is 30.
            db_batch_size (int): The number of records to insert into the database at once if it exists. Default is 100.
        """
        # Set the conversion options
        self.axial_only = axial_only
        self.no_mip = no_mip
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        self.min_slice_count = min_slice_count
        self.db_batch_size = db_batch_size
        self.dbase_config = read_dbase_config()
        # Set the resampling options for the dicom2nifti library
        settings.resample = resample
        settings.resample_spline_interpolation_order = (
            resample_spline_interpolation_order
        )
        settings.resample_padding = resample_padding
        settings.validate_orthogonal = validate_orthogonal
        settings.validate_slice_increment = validate_slice_increment
        settings.validate_instance_number = validate_instance_number
        settings.validate_slicecount = validate_slice_count

    def convert(
        self, dicom_folders: list[str | Path], output_directory: str | Path
    ) -> None:
        """
        Convert DICOM files to NIfTI format.

        Args:
            dicom_folders (list[str|Path]): A list of DICOM folders to convert.
            output_directory (str|Path): Directory to save the converted NIfTI files.
        """
        # Validate the input dicom folders
        if not isinstance(dicom_folders, list):
            raise TypeError("dicom_folders must be a list of folder paths.")
        num_folders = len(dicom_folders)
        if num_folders == 0:
            raise ValueError("dicom_folders list is empty.")
        # Create the base output directory if it doesn't exist
        self.output_directory = Path(output_directory).resolve()
        if not self.output_directory.exists():
            logger.info(f"Creating output directory: {self.output_directory}")
            self.output_directory.mkdir(parents=True, exist_ok=True)
        # Run the conversion in parallel using multiprocessing
        if self.n_processes > num_folders:
            logger.info(
                f"Number of processes ({self.n_processes}) is greater than number of folders ({num_folders}). Setting number of processes to {num_folders}."
            )
            self.n_processes = num_folders
        logger.info(
            f"Running in {self.n_processes} process{'es' if self.n_processes > 1 else ''}."
        )
        completed = 1
        metadata_batch = []
        with Pool(self.n_processes) as pool:
            # We use imap_unordered because we don't care about the order of completion
            for result in pool.imap_unordered(self.convert_folder, dicom_folders):
                if (result is not None) and (len(result) == 3):
                    dicom_folder, output_nifti, metadata = result
                    metadata["output_nifti"] = str(output_nifti)
                    metadata["conversion_date"] = date.today().strftime("%Y-%m-%d")
                    logger.success(
                        f"{completed}/{num_folders} - Converted {dicom_folder} to {output_nifti}",
                        extra={"folder": dicom_folder, "output": output_nifti},
                    )
                    if self.dbase_config and metadata:
                        metadata_batch.append(metadata)
                        if len(metadata_batch) >= self.db_batch_size:
                            insert_data_batch(
                                self.dbase_config,
                                "conversions",
                                metadata_batch,
                            )
                            metadata_batch = []
                else:
                    logger.error(
                        f"{completed}/{num_folders} - Failed to convert {dicom_folder}."
                    )
                completed += 1
        # Insert any remaining metadata into the database
        if self.dbase_config and metadata_batch:
            insert_data_batch(
                self.dbase_config,
                "conversions",
                metadata_batch,
            )

    def convert_folder(self, dicom_folder: str | Path) -> tuple[str, str] | None:
        """
        Wrapper of _convert_folder to handle exceptions and logging.
        Args:
            dicom_folder (str): Path to the DICOM folder to convert.
        Returns:
            tuple[str, str]: The input DICOM folder and output nifti file path.
        """
        try:
            return self._convert_folder(dicom_folder)
        except Exception as e:
            logger.error(
                f"Error converting folder {dicom_folder}: {e}\n{traceback.format_exc()}",
                extra={"folder": dicom_folder, "error": str(e)},
            )
            return None

    def _convert_folder(self, dicom_folder: str | Path) -> tuple[str, str]:
        """
        Convert a single DICOM folder to NIfTI format.

        Args:
            dicom_folder (str): Path to the DICOM folder to convert.
        Returns:
            tuple[str, str]: The input DICOM folder and output nifti file path.
        """
        with threadpool_limits(limits=self.threads_per_process):
            dicom_folder = Path(dicom_folder).resolve()
            if not dicom_folder.exists():
                raise FileNotFoundError(f"DICOM folder does not exist: {dicom_folder}")
            if not dicom_folder.is_dir():
                raise NotADirectoryError(f"Path is not a directory: {dicom_folder}")
            dicom_files = list(dicom_folder.glob("*.dcm"))
            if len(dicom_files) == 0:
                logger.warning(f"No DICOM files found in folder: {dicom_folder}")
                return
            if len(dicom_files) < self.min_slice_count:
                logger.warning(
                    f"Skipping folder {dicom_folder} with less than {self.min_slice_count} DICOM files."
                )
                return
            # Get the metadata from the DICOM files
            reference_dicom = dicom_files[0]
            metadata = self._get_metadata(reference_dicom)
            if metadata.get("series_name") == "nan":
                logger.warning(
                    f"Skipping folder {dicom_folder} due to series_name == 'nan'."
                )
                return
            if "ai-rad_companion" in metadata.get("series_name"):
                logger.warning(
                    f"Skipping AI-Rad companion report series {metadata.get('original_series_name')} in folder {dicom_folder}."
                )
                return
            if self.axial_only and metadata.get("ct_direction") != "AX":
                logger.warning(
                    f"Skipping non-axial series {metadata.get('original_series_name')} in folder {dicom_folder}."
                )
                return
            if self.no_mip:
                # Do a check for MIP series
                if (
                    "mip" in metadata.get("series_name").lower()
                    or metadata.get("slice_thickness_mm", 5) > 5
                ):
                    logger.warning(
                        f"Skipping likely MIP series {metadata.get('original_series_name')} in folder {dicom_folder}."
                    )
                    return
            output_nifti = (
                f"{metadata['series_number']}_{metadata['series_name']}.nii.gz"
            )
            custom_output_directory = (
                self.output_directory
                / metadata.get("mrn")[:2]  # First 2 numbers of MRN
                / metadata.get("mrn")[2:4]  # Next 2 numbers of MRN
                / metadata.get("mrn")
                / metadata.get("accession")
                / metadata.get("series_name")
            )
            if not custom_output_directory.exists():
                custom_output_directory.mkdir(parents=True, exist_ok=True)
            try:
                # Convert the DICOM files to NIfTI format
                dicom2nifti.convert_directory(dicom_folder, custom_output_directory)
                # Save the metadata to a JSON file
                metadata_file = custom_output_directory / "metadata.json"
                with metadata_file.open("w") as f:
                    json.dump(metadata, f, indent=4)
                # Copy the reference DICOM file to the output directory
                dicom_file = custom_output_directory / reference_dicom.name
                copyfile(reference_dicom, dicom_file)
                return dicom_folder, custom_output_directory / output_nifti, metadata
            except ConversionError as e:
                logger.warning("Failed to convert dicom {}: {}", dicom_folder, e)
                return None

    def _get_metadata(self, reference_dicom: Path) -> dict:
        """Get the metadata from a DICOM file that is not saved in the NIfTI header.
        Args:
            reference_dicom (Path): Path to the reference DICOM file.
        Returns:
            dict: A dictionary containing the metadata.
        """
        orientations = {
            "COR": [1, 0, 0, 0, 0, -1],
            "SAG": [0, 1, 0, 0, 0, -1],
            "AX": [1, 0, 0, 0, 1, 0],
        }
        meta_attributes = {
            "PatientID": "mrn",
            "AccessionNumber": "accession",
            "SeriesDescription": "series_name",
            "SeriesNumber": "series_number",
            "StudyDescription": "study_description",
            "ImageOrientationPatient": "ct_direction",
            "ImageType": "image_type",
            "PatientSex": "sex",
            "PatientAge": "age",
            "PatientSize": "height_m",
            "PatientWeight": "weight_kg",
            "PregnancyStatus": "pregnancy_status",
            "PatientBirthDate": "birth_date",
            "AcquisitionDate": "scan_date",
            "PixelSpacing": None,
            "SliceThickness": "slice_thickness_mm",
            "Manufacturer": "manufacturer",
            "ManufacturerModelName": "model",
            "KVP": "kvp",
            "SequenceName": "sequence_name",
            "ProtocolName": "protocol_name",
            "ContrastBolusAgent": "contrast_bolus_agent",
            "ContrastBolusRoute": "contrast_bolus_route",
            "ContrastBolusVolume": "contrast_bolus_volume",
            "StudyInstanceUID": "study_uid",
            "SeriesInstanceUID": "series_uid",
            "Modality": "modality",
        }
        metadata = {}
        try:
            ref = dcmread(reference_dicom)
        except FileNotFoundError:
            logger.error(f"Reference DICOM file not found: {reference_dicom}")
            return metadata
        except Exception as e:
            logger.error(f"Error reading DICOM file {reference_dicom}: {e}")
            return metadata
        # Need these 3 fields to create the output directory
        if not hasattr(ref, "PatientID"):
            raise ValueError(
                f"Reference DICOM file does not have PatientID: {reference_dicom.parent}"
            )
        if not hasattr(ref, "AccessionNumber"):
            raise ValueError(
                f"Reference DICOM file does not have AccessionNumber: {reference_dicom.parent}"
            )
        if not hasattr(ref, "SeriesDescription"):
            raise ValueError(
                f"Reference DICOM file does not have SeriesDescription: {reference_dicom.parent}"
            )
        # Go through the metadata attributes and get the values
        for tag, column in meta_attributes.items():
            if not hasattr(ref, tag):
                # Not every DICOM file has all the tags, so we skip them
                continue
            value = getattr(ref, tag)
            match tag:
                case "PatientID":
                    value = str(value).zfill(0)
                case "SeriesDescription":
                    # Remove accents from the SeriesDescription
                    metadata["original_series_name"] = value
                    value = _remove_accents(value)
                case "SeriesNumber":
                    value = int(value)
                case "ImageOrientationPatient":
                    value = [round(float(x)) for x in value]
                    for orientation, array in orientations.items():
                        if np.allclose(value, array):
                            value = orientation
                            break
                    if isinstance(value, list):
                        value = "AX"
                    if isinstance(value, str) and value not in orientations:
                        value = "AX"
                case "ImageType":
                    value = "_".join(value)
                case tag if tag in ["PatientBirthDate", "AcquisitionDate"]:
                    value = date.fromisoformat(value).strftime("%Y-%m-%d")
                case "PatientAge":
                    value = int(value[:-1])
                case "PixelSpacing":
                    length, width = value
                    metadata["pixel_length_mm"] = float(length)
                    metadata["pixel_width_mm"] = float(width)
                    continue
                case "PregnancyStatus":
                    value = int(value)
                    if value == 4:
                        value = None
                    elif value == 1:
                        value = 0
                    elif value == 2:
                        value = 1
                    elif value == 3:
                        value = 2
                case tag if tag in [
                    "PatientSize",
                    "PatientWeight",
                    "KVP",
                    "SliceThickness",
                ]:
                    value = float(value)
                case tag if tag in ["SequenceName", "ProtocolName"]:
                    value = str(value).lower()
                case _:
                    pass
            if value == "":
                value = None
            metadata[column] = value
        if not metadata:
            raise ValueError(
                f"Reference DICOM file does not have any metadata: {reference_dicom.parent}"
            )
        if metadata.get("slice_thickness_mm") is None:
            metadata["slice_thickness_mm"] = 5
        metadata["dicom_folder"] = str(reference_dicom.parent)
        return metadata
